"""
Calculate the period between two consecutive waves for each wave and channel.
"""

import argparse
from pathlib import Path
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_path, none_or_str
from utils.neo_utils import analogsignal_to_imagesequence

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_path, default=None,
                 help="path of output image file")
CLI.add_argument("--kernel", "--KERNEL", nargs='?', type=none_or_str, default=None,
                 help="derivative kernel")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")

def calc_local_wave_intervals(evts):
    wave_labels = evts.labels.astype(int)
    unique_labels = np.sort(np.unique(wave_labels))
    unique_channels = np.sort(np.unique(evts.array_annotations['channels'].astype(int)))

    channel_idx_map = np.empty(np.max(unique_channels)+1) * np.nan
    for i, channel in enumerate(unique_channels):
        channel_idx_map[channel] = i

    trigger_collection = np.empty((len(unique_labels), len(unique_channels)),
                                  dtype=float) * np.nan

    for (i, wave_id) in enumerate(unique_labels):
        wave_trigger_evts = evts[wave_labels == wave_id]

        channels = wave_trigger_evts.array_annotations['channels'].astype(int)

        channel_idx = channel_idx_map[channels].astype(int)
        trigger_collection[i, channel_idx] = wave_trigger_evts.times

    intervals = np.diff(trigger_collection, axis=0)
    intervals = intervals.reshape((len(unique_labels)-1)*len(unique_channels))

    mask = np.isfinite(intervals)
    intervals = intervals[mask]

    channel_ids = np.tile(unique_channels, len(unique_labels)-1)[mask]
    wave_ids = np.repeat(unique_labels[:-1], len(unique_channels))[mask]

    return wave_ids, channel_ids, intervals*evts.times.units



if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]
    imgseq = analogsignal_to_imagesequence(asig)

    evts = block.filter(name=args.event_name, objects="Event")[0]
    evts = evts[evts.labels != '-1']

    wave_ids, channel_ids, intervals = calc_local_wave_intervals(evts)

    # transform to DataFrame
    df = pd.DataFrame(intervals.magnitude, columns=['inter_wave_interval_local'])
    df['inter_wave_interval_local_unit'] = [intervals.dimensionality.string]*len(channel_ids)
    df['channel_id'] = channel_ids
    df[f'{args.event_name}_id'] = wave_ids

    df.to_csv(args.output)

    fig, ax = plt.subplots()
    ax.hist(1./intervals.magnitude[np.where(np.isfinite(1./intervals))[0]],
            bins=100, range=[0, 8])
    plt.xlabel('local rate of waves (Hz)', fontsize=7.)
    if args.output_img is not None:
        save_plot(args.output_img)



"""
Create a plot of the channel-wise and average power spectrum density.
"""

import numpy as np
import quantities as pq
import argparse
from pathlib import Path
import matplotlib.pyplot as plt
import seaborn as sns
from elephant.spectral import welch_psd
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_float

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output figure")
CLI.add_argument("--highpass_frequency", nargs='?', type=none_or_float,
                 default='None', help="lower bound of frequency band in Hz")
CLI.add_argument("--lowpass_frequency", nargs='?', type=none_or_float,
                 default='None', help="upper bound of frequency band in Hz")
CLI.add_argument("--psd_frequency_resolution", nargs='?', type=float, default=5,
                 help="frequency resolution of the power spectrum in Hz")
CLI.add_argument("--psd_overlap", nargs='?', type=float, default=0.5,
                 help="overlap parameter for Welch's algorithm [0-1]")

def plot_psd(frequencies, psd, highpass_frequency, lowpass_frequency):
    sns.set(style='ticks', palette="deep", context="notebook")
    fig, ax = plt.subplots()
    for channel in psd:
        ax.semilogy(frequencies, channel, alpha=0.7)
    ax.semilogy(frequencies, np.mean(psd, axis=0), linewidth=2,
                color='k', label='channel average')
    ax.set_title('Power Spectrum')
    ax.set_xlabel('frequency [Hz]')
    ax.set_ylabel('power spectral density')

    if highpass_frequency is not None and lowpass_frequency is not None:
        left = highpass_frequency if highpass_frequency is not None else 0
        right = lowpass_frequency if lowpass_frequency is not None else ax.get_xlim()[1]
        ax.axvspan(left, right, alpha=0.2, color='k', label='filtered region')

    plt.legend()
    return fig


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    asig = load_neo(args.data, 'analogsignal')

    freqs, psd = welch_psd(asig,
                           frequency_resolution=args.psd_frequency_resolution*pq.Hz,
                           window='hann',
                           overlap=args.psd_overlap)

    plot_psd(frequencies=freqs,
             psd=psd,
             highpass_frequency=args.highpass_frequency,
             lowpass_frequency=args.lowpass_frequency)

    save_plot(args.output)



"""
Calculate the spatial derivative on the time-delays of the wave triggers
in each channel.

The derivative is calculated using a kernel convolution.
"""

import argparse
from pathlib import Path
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.interpolate import RBFInterpolator
from utils.convolve import get_kernel, nan_conv2d
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_path, none_or_str, str_to_bool

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_path, default=None,
                 help="path of output image file")
CLI.add_argument("--kernel", "--KERNEL", nargs='?', type=none_or_str, default=None,
                 help="derivative kernel")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")
CLI.add_argument("--interpolate", "--INTERPOLATE", nargs='?', type=str_to_bool, default=False,
                 help="whether to thin-plate-spline interpolate the wave patterns before derivation")
CLI.add_argument("--smoothing", "--SMOOTHING", nargs='?', type=float, default=0,
                 help="smoothing factor for the interpolation")

def interpolate_grid(grid, smoothing):
    y, x = np.where(np.isfinite(grid))
    rbf_func = RBFInterpolator(np.stack((y,x), axis=-1),
                               grid[y,x],
                               neighbors=None,
                               smoothing=smoothing,
                               kernel='thin_plate_spline',
                               epsilon=None,
                               degree=None)
    return rbf_func

def sample_wave_pattern(pattern_func, dim_x, dim_y):
    fy, fx = np.meshgrid(np.arange(0, dim_y),
                         np.arange(0, dim_x),
                         indexing='ij')
    fcoords = np.stack((fy,fx), axis=-1)
    wave_pattern = pattern_func(fcoords.reshape(-1,2))
    return wave_pattern.reshape(dim_y, dim_x)

def calc_spatial_derivative(evts, kernel_name, interpolate=False, smoothing=0):
    labels = evts.labels.astype(int)
    dim_x = int(max(evts.array_annotations['x_coords']))+1
    dim_y = int(max(evts.array_annotations['y_coords']))+1

    spatial_derivative_df = pd.DataFrame()

    for wave_id in np.unique(labels):
        wave_trigger_evts = evts[labels == wave_id]

        x_coords = wave_trigger_evts.array_annotations['x_coords'].astype(int)
        y_coords = wave_trigger_evts.array_annotations['y_coords'].astype(int)
        channels = wave_trigger_evts.array_annotations['channels'].astype(int)

        trigger_collection = np.empty([dim_y, dim_x]) * np.nan
        trigger_collection[y_coords, x_coords] = wave_trigger_evts.times

        if interpolate:
            try:
                pattern_func = interpolate_grid(trigger_collection, smoothing)
                trigger_collection = sample_wave_pattern(pattern_func,
                                                         dim_x=dim_x, dim_y=dim_y)
            except ValueError as ve:
                warn(repr(ve))
                warn('Continuing without interpolation.')

        kernel = get_kernel(kernel_name)
        d_horizont = -1 * nan_conv2d(trigger_collection, kernel.x) # is -1 correct?
        d_vertical = -1 * nan_conv2d(trigger_collection, kernel.y)

        dt_y = d_vertical[y_coords, x_coords]
        dt_x = d_horizont[y_coords, x_coords]

        df = pd.DataFrame(list(zip(dt_x, dt_y, x_coords, y_coords, channels)),
                          columns=['dt_x', 'dt_y', 'x_coords', 'y_coords', 'channel_id'])
        df[f'{evts.name}_id'] = wave_id

        spatial_derivative_df = pd.concat([spatial_derivative_df, df],
                                          ignore_index=True)

    fig, ax = plt.subplots(ncols=3, figsize=(15,5))
    img = ax[0].imshow(trigger_collection, cmap='viridis', origin='lower')
    plt.colorbar(img, ax=ax[0])
    vminmax = np.nanmax(abs(d_vertical))
    img = ax[1].imshow(d_vertical, origin='lower', cmap='coolwarm',
                       vmin=-vminmax, vmax=vminmax)
    plt.colorbar(img, ax=ax[1])
    vminmax = np.nanmax(abs(d_horizont))
    img = ax[2].imshow(d_horizont, origin='lower', cmap='coolwarm',
                       vmin=-vminmax, vmax=vminmax)
    plt.colorbar(img, ax=ax[2])
    ax[0].set_title(f'wave {wave_id}')
    ax[1].set_title('dt Y (vertical)')
    ax[2].set_title('dt X (horizontal)')
    ax[0].set_axis_off()
    ax[1].set_axis_off()
    ax[2].set_axis_off()

    return spatial_derivative_df


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = block.segments[0].analogsignals[0]
    evts = block.filter(name=args.event_name, objects="Event")[0]
    evts = evts[evts.labels != '-1']

    df = calc_spatial_derivative(evts,
                                 kernel_name=args.kernel,
                                 interpolate=args.interpolate,
                                 smoothing=args.smoothing)

    df['kernel'] = args.kernel
    df['spatial_scale'] = evts.annotations['spatial_scale'].magnitude
    df['spatial_scale_unit'] = evts.annotations['spatial_scale'].dimensionality.string
    df['dt_unit'] = evts.times.dimensionality.string

    df.to_csv(args.output)

    if args.output_img is not None:
        save_plot(args.output_img)



"""
Plot snapshots of the input data showing the detected waves.
"""

import os
import numpy as np
import quantities as pq
import argparse
from pathlib import Path
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.neo_utils import analogsignal_to_imagesequence

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path,
                 help="path to input data in neo format")
CLI.add_argument("--output_dir", nargs='?', type=Path,
                 help="path to output directory")
CLI.add_argument("--img_name", nargs='?', type=str,
                 help="")
CLI.add_argument("--time_window", nargs='?', type=float, default=0.4,
                 help="size of the plotted window in seconds.")
CLI.add_argument("--colormap", nargs='?', type=str, default='viridis',
                 help="")

def plot_wave(wave_id, waves_event, asig, frames, vec_frames,
              time_window=0.4*pq.s, cmap='virids'):
    idx = np.where(waves_event.labels == str(wave_id))[0]
    x_coords = waves_event.array_annotations['x_coords'][idx]
    y_coords = waves_event.array_annotations['y_coords'][idx]

    t = waves_event.times[idx]
    time_steps = np.unique(t)

    dim_t, dim_y, dim_x = frames.shape
    y_idx, x_idx = np.meshgrid(np.arange(dim_y), np.arange(dim_x), indexing='ij')
    markersize = 50 / max([dim_x, dim_y])
    skip_step = int(min([dim_x, dim_y]) / 50) + 1

    vmin, vmax = np.nanmin(frames), np.nanmax(frames)

    time_window_steps = asig.sampling_rate.rescale('Hz') * time_window.rescale('s')
    half_window = int(time_window_steps.magnitude / 2)

    fig, axes = plt.subplots(nrows=2, ncols=len(time_steps),
                             figsize=(2*len(time_steps), 5), sharey='row')
    if len(time_steps) < 2:
        axes = [[axes[0]], [axes[1]]]

    axes[0][0].set_ylabel(f"Wave {wave_id}", fontsize=20)

    for i, ax in enumerate(axes[0]):
        ax.set_title('{:.3f} '.format(time_steps[i].magnitude) + f'{time_steps[i].dimensionality}')

    for i in idx:
        x = waves_event.array_annotations['x_coords'][i]
        y = waves_event.array_annotations['y_coords'][i]
        t = waves_event.times[i]
        ax_i = np.where(time_steps == t)[0][0]

        channel = waves_event.array_annotations['channels'][i]
        t_i = np.argmax(asig.times >= t)
        i_start = np.max([0, t_i-half_window])
        i_stop = np.min([t_i+half_window, len(asig.times)-1])
        axes[0][ax_i].plot(asig.times[i_start : i_stop],
                           asig.as_array()[i_start : i_stop, channel])
        axes[0][ax_i].axvline(t, color='r')

        axes[1][ax_i].imshow(frames[t_i], origin='lower', cmap=cmap,
                             vmin=vmin, vmax=vmax)
        axes[1][ax_i].plot(x, y, linestyle='None', marker='D',
                           markersize=markersize, color='r')
        axes[1][ax_i].set_axis_off()

        axes[1][ax_i].quiver(x_idx[::skip_step,::skip_step],
                             y_idx[::skip_step,::skip_step],
                             np.real(vec_frames[t_i])[::skip_step,::skip_step],
                             np.imag(vec_frames[t_i])[::skip_step,::skip_step],
                             zorder=5)
    return axes


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = block.segments[0].analogsignals[0]
    vec_asig = block.filter(name='optical_flow', objects="AnalogSignal")[0]

    frames = analogsignal_to_imagesequence(asig).as_array()
    vec_frames = analogsignal_to_imagesequence(vec_asig).as_array()

    waves_event = block.filter(name='wavefronts', objects="Event")[0]

    cmap = plt.get_cmap(args.colormap)

    for wave_id in np.unique(waves_event.labels):
        if int(wave_id) != -1:  # collection of not-clustered triggers
            ax = plot_wave(wave_id=wave_id,
                           waves_event=waves_event,
                           asig=asig,
                           frames=frames,
                           vec_frames=vec_frames,
                           time_window=args.time_window*pq.s,
                           cmap=cmap)

            output_path = os.path.join(args.output_dir,
                                       args.img_name.replace('id0', f'id{wave_id}'))
            save_plot(output_path)
            plt.close()



"""
Calculate the wave propagation velocity for each wave and channel.
"""

import argparse
from pathlib import Path
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_path

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to spatial derivative dataframe")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_path, default=None,
                 help="path of output image file")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    df = pd.read_csv(args.data)

    velocity = df.spatial_scale * np.sqrt(1/(df.dt_x**2 + df.dt_y**2))
    velocity[~np.isfinite(velocity)] = np.nan

    velocity_df = pd.DataFrame(velocity, columns=['velocity_local'])
    velocity_df['channel_id'] = df.channel_id
    velocity_df['velocity_local_unit'] = f'{df.spatial_scale_unit[0]}/{df.dt_unit[0]}'
    velocity_df[f'{args.event_name}_id'] = df[f'{args.event_name}_id']

    velocity_df.to_csv(args.output)

    if args.output_img is not None:
        save_plot(args.output_img)



"""
Plot an example signal trace before and after application of some processing steps.
"""

import matplotlib.pyplot as plt
import seaborn as sns
import argparse
from pathlib import Path
import os
from utils.io_utils import load_neo, save_plot
from utils.neo_utils import time_slice
from utils.parse import none_or_float

CLI = argparse.ArgumentParser()
CLI.add_argument("--original_data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--img_dir", nargs='?', type=Path, required=True,
                 help="path of output figure directory")
CLI.add_argument("--img_name", nargs='?', type=str,
                 default='processed_trace_channel0.png',
                 help='example filename for channel 0')
CLI.add_argument("--t_start", nargs='?', type=none_or_float, default=0,
                 help="start time in seconds")
CLI.add_argument("--t_stop", nargs='?', type=none_or_float, default=10,
                 help="stop time in seconds")
CLI.add_argument("--channels", nargs='+', type=int, default=0,
                 help="channel to plot")

def plot_traces(original_asig, processed_asig, channel):
    sns.set(style='ticks', palette="deep", context="notebook")
    fig, ax1 = plt.subplots()
    palette = sns.color_palette()

    ax1.plot(original_asig.times,
            original_asig.as_array()[:,channel],
            color=palette[0])
    ax1.set_ylabel('original signal', color=palette[0])
    ax1.tick_params('y', colors=palette[0])

    ax2 = ax1.twinx()
    ax2.plot(processed_asig.times,
            processed_asig.as_array()[:,channel],
            color=palette[1])
    ax2.set_ylabel('processed signal', color=palette[1])
    ax2.tick_params('y', colors=palette[1])

    ax1.set_title('Channel {}'.format(channel))
    ax1.set_xlabel('time [{}]'.format(original_asig.times.units.dimensionality.string))

    return ax1, ax2


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    orig_asig = load_neo(args.original_data, 'analogsignal', lazy=False)
    orig_asig = time_slice(orig_asig, t_start=args.t_start, t_stop=args.t_stop,
                           lazy=False, channel_indexes=args.channels)

    proc_asig = load_neo(args.data, 'analogsignal', lazy=False)
    proc_asig = time_slice(proc_asig, t_start=args.t_start, t_stop=args.t_stop,
                           lazy=False, channel_indexes=args.channels)

    for channel in args.channels:
        plot_traces(orig_asig, proc_asig, channel)
        output_path = os.path.join(args.img_dir,
                                   args.img_name.replace('_channel0', f'_channel{channel}'))
        save_plot(output_path)



"""
Plot the critical points in the optical flow vector field.
"""

import argparse
from pathlib import Path
import numpy as np
from copy import copy
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.neo_utils import analogsignal_to_imagesequence

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--skip_step", nargs='?', type=int, default=3,
                 help="skipping every x vector for the plot")
CLI.add_argument("--frame_id", nargs='?', type=int, default=0,
                 help="number of the frame to plot")

def plot_frame(frame, ax=None, skip_step=3):
    dim_y, dim_x = frame.shape

    if ax is None:
        fig, ax = plt.subplots()

    ax.quiver(np.arange(dim_x)[::skip_step],
              np.arange(dim_y)[::skip_step],
              np.real(frame[::skip_step,::skip_step]),
              np.imag(frame[::skip_step,::skip_step]))

    Y, X = np.meshgrid(np.arange(dim_y), np.arange(dim_x), indexing='ij')
    ZR = np.real(frame)
    ZI = np.imag(frame)
    contourR = ax.contour(X, Y, ZR, levels=[0], colors='b', label='x = 0')
    contourI = ax.contour(X, Y, ZI, levels=[0], colors='g', label='y = 0')
    return ax


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()
    block = load_neo(args.data)

    asigs = block.filter(name='optical_flow', objects="AnalogSignal")

    if asigs:
        asig = asigs[0]
    else:
        raise ValueError("Input does not contain a signal with name " \
                       + "'optical_flow'!")

    imgseq = analogsignal_to_imagesequence(asig)

    crit_point_evt = [evt for evt in block.segments[0].events
                      if evt.name == "critical_points"]
    if crit_point_evt:
        crit_point_evt = crit_point_evt[0]
    else:
        raise ValueError("Input does not contain an event with name " \
                       + "'critical_points'!")

    fig, ax = plt.subplots()

    ax = plot_frame(imgseq.as_array()[args.frame_id],
                    skip_step=args.skip_step,
                    ax=ax)

    start_id = np.argmax(crit_point_evt.times >= asig.times[args.frame_id])
    stop_id = np.argmax(crit_point_evt.times >= asig.times[args.frame_id+1])

    ax.scatter(crit_point_evt.array_annotations['x'][start_id:stop_id],
               crit_point_evt.array_annotations['y'][start_id:stop_id],
               color='r')

    ax.set_title(f"{asig.times[args.frame_id]:.2f}")
    ax.set_xticklabels([])
    ax.set_yticklabels([])

    save_plot(args.output)
